"""
Advanced Streaming Format (ASF) parser, format used by Windows Media Video
(WMF) and Windows Media Audio (WMA).

Informations:
- http://www.microsoft.com/windows/windowsmedia/forpros/format/asfspec.aspx
- http://swpat.ffii.org/pikta/xrani/asf/index.fr.html

Author: Victor Stinner
Creation: 5 august 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, ParserError,
                           UInt16, UInt32, UInt64,
                           TimestampWin64, TimedeltaWin64,
                           TimedeltaMillisWin64,
                           String, PascalString16, Enum,
                           Bit, Bits, PaddingBits,
                           PaddingBytes, NullBytes, RawBytes)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import (
    displayHandler, filesizeHandler)
from hachoir.core.tools import humanBitRate

from hachoir.parser.video.fourcc import audio_codec_name, video_fourcc_name
from hachoir.parser.common.win32 import BitmapInfoHeader, GUID

MAX_HEADER_SIZE = 100 * 1024  # bytes


class AudioHeader(FieldSet):
    guid = "F8699E40-5B4D-11CF-A8FD-00805F5C442B"

    def createFields(self):
        yield Enum(UInt16(self, "twocc"), audio_codec_name)
        yield UInt16(self, "channels")
        yield UInt32(self, "sample_rate")
        yield UInt32(self, "bit_rate")
        yield UInt16(self, "block_align")
        yield UInt16(self, "bits_per_sample")
        yield UInt16(self, "codec_specific_size")
        size = self["codec_specific_size"].value
        if size:
            yield RawBytes(self, "codec_specific", size)


class BitrateMutualExclusion(FieldSet):
    guid = "D6E229DC-35DA-11D1-9034-00A0C90349BE"
    mutex_name = {
        "D6E22A00-35DA-11D1-9034-00A0C90349BE": "Language",
        "D6E22A01-35DA-11D1-9034-00A0C90349BE": "Bitrate",
        "D6E22A02-35DA-11D1-9034-00A0C90349BE": "Unknown",
    }

    def createFields(self):
        yield Enum(GUID(self, "exclusion_type"), self.mutex_name)
        yield UInt16(self, "nb_stream")
        for index in range(self["nb_stream"].value):
            yield UInt16(self, "stream[]")


class VideoHeader(FieldSet):
    guid = "BC19EFC0-5B4D-11CF-A8FD-00805F5C442B"

    def createFields(self):
        if False:
            yield UInt32(self, "width0")
            yield UInt32(self, "height0")
            yield PaddingBytes(self, "reserved[]", 7)
            yield UInt32(self, "width")
            yield UInt32(self, "height")
            yield PaddingBytes(self, "reserved[]", 2)
            yield UInt16(self, "depth")
            yield Enum(String(self, "codec", 4, charset="ASCII"), video_fourcc_name)
            yield NullBytes(self, "padding", 20)
        else:
            yield UInt32(self, "width")
            yield UInt32(self, "height")
            yield PaddingBytes(self, "reserved[]", 1)
            yield UInt16(self, "format_data_size")
            if self["format_data_size"].value < 40:
                raise ParserError("Unknown format data size")
            yield BitmapInfoHeader(self, "bmp_info", use_fourcc=True)


class FileProperty(FieldSet):
    guid = "8CABDCA1-A947-11CF-8EE4-00C00C205365"

    def createFields(self):
        yield GUID(self, "guid")
        yield filesizeHandler(UInt64(self, "file_size"))
        yield TimestampWin64(self, "creation_date")
        yield UInt64(self, "pckt_count")
        yield TimedeltaWin64(self, "play_duration")
        yield TimedeltaWin64(self, "send_duration")
        yield TimedeltaMillisWin64(self, "preroll")
        yield Bit(self, "broadcast", "Is broadcast?")
        yield Bit(self, "seekable", "Seekable stream?")
        yield PaddingBits(self, "reserved[]", 30)
        yield filesizeHandler(UInt32(self, "min_pckt_size"))
        yield filesizeHandler(UInt32(self, "max_pckt_size"))
        yield displayHandler(UInt32(self, "max_bitrate"), humanBitRate)


class HeaderExtension(FieldSet):
    guid = "5FBF03B5-A92E-11CF-8EE3-00C00C205365"

    def createFields(self):
        yield GUID(self, "reserved[]")
        yield UInt16(self, "reserved[]")
        yield UInt32(self, "size")
        if self["size"].value:
            yield RawBytes(self, "data", self["size"].value)


class Header(FieldSet):
    guid = "75B22630-668E-11CF-A6D9-00AA0062CE6C"

    def createFields(self):
        yield UInt32(self, "obj_count")
        yield PaddingBytes(self, "reserved[]", 2)
        for index in range(self["obj_count"].value):
            yield Object(self, "object[]")


class Metadata(FieldSet):
    guid = "75B22633-668E-11CF-A6D9-00AA0062CE6C"
    names = ("title", "author", "copyright", "xxx", "yyy")

    def createFields(self):
        for index in range(5):
            yield UInt16(self, "size[]")
        for name, size in zip(self.names, self.array("size")):
            if size.value:
                yield String(self, name, size.value, charset="UTF-16-LE", strip=" \0")


class Descriptor(FieldSet):
    """
    See ExtendedContentDescription class.
    """
    TYPE_BYTE_ARRAY = 1
    TYPE_NAME = {
        0: "Unicode",
        1: "Byte array",
        2: "BOOL (32 bits)",
        3: "DWORD (32 bits)",
        4: "QWORD (64 bits)",
        5: "WORD (16 bits)"
    }

    def createFields(self):
        yield PascalString16(self, "name", "Name", charset="UTF-16-LE", strip="\0")
        yield Enum(UInt16(self, "type"), self.TYPE_NAME)
        yield UInt16(self, "value_length")
        type = self["type"].value
        size = self["value_length"].value
        name = "value"
        if type == 0 and (size % 2) == 0:
            yield String(self, name, size, charset="UTF-16-LE", strip="\0")
        elif type in (2, 3):
            yield UInt32(self, name)
        elif type == 4:
            yield UInt64(self, name)
        else:
            yield RawBytes(self, name, size)


class ExtendedContentDescription(FieldSet):
    guid = "D2D0A440-E307-11D2-97F0-00A0C95EA850"

    def createFields(self):
        yield UInt16(self, "count")
        for index in range(self["count"].value):
            yield Descriptor(self, "descriptor[]")


class Codec(FieldSet):
    """
    See CodecList class.
    """
    type_name = {
        1: "video",
        2: "audio"
    }

    def createFields(self):
        yield Enum(UInt16(self, "type"), self.type_name)
        yield UInt16(self, "name_len", "Name length in character (byte=len*2)")
        if self["name_len"].value:
            yield String(self, "name", self["name_len"].value * 2, "Name", charset="UTF-16-LE", strip=" \0")
        yield UInt16(self, "desc_len", "Description length in character (byte=len*2)")
        if self["desc_len"].value:
            yield String(self, "desc", self["desc_len"].value * 2, "Description", charset="UTF-16-LE", strip=" \0")
        yield UInt16(self, "info_len")
        if self["info_len"].value:
            yield RawBytes(self, "info", self["info_len"].value)


class CodecList(FieldSet):
    guid = "86D15240-311D-11D0-A3A4-00A0C90348F6"

    def createFields(self):
        yield GUID(self, "reserved[]")
        yield UInt32(self, "count")
        for index in range(self["count"].value):
            yield Codec(self, "codec[]")


class SimpleIndexEntry(FieldSet):
    """
    See SimpleIndex class.
    """

    def createFields(self):
        yield UInt32(self, "pckt_number")
        yield UInt16(self, "pckt_count")


class SimpleIndex(FieldSet):
    guid = "33000890-E5B1-11CF-89F4-00A0C90349CB"

    def createFields(self):
        yield GUID(self, "file_id")
        yield TimedeltaWin64(self, "entry_interval")
        yield UInt32(self, "max_pckt_count")
        yield UInt32(self, "entry_count")
        for index in range(self["entry_count"].value):
            yield SimpleIndexEntry(self, "entry[]")


class BitRate(FieldSet):
    """
    See BitRateList class.
    """

    def createFields(self):
        yield Bits(self, "stream_index", 7)
        yield PaddingBits(self, "reserved", 9)
        yield displayHandler(UInt32(self, "avg_bitrate"), humanBitRate)


class BitRateList(FieldSet):
    guid = "7BF875CE-468D-11D1-8D82-006097C9A2B2"

    def createFields(self):
        yield UInt16(self, "count")
        for index in range(self["count"].value):
            yield BitRate(self, "bit_rate[]")


class Data(FieldSet):
    guid = "75B22636-668E-11CF-A6D9-00AA0062CE6C"

    def createFields(self):
        yield GUID(self, "file_id")
        yield UInt64(self, "packet_count")
        yield PaddingBytes(self, "reserved", 2)
        size = (self.size - self.current_size) // 8
        yield RawBytes(self, "data", size)


class StreamProperty(FieldSet):
    guid = "B7DC0791-A9B7-11CF-8EE6-00C00C205365"

    def createFields(self):
        yield GUID(self, "type")
        yield GUID(self, "error_correction")
        yield UInt64(self, "time_offset")
        yield UInt32(self, "data_len")
        yield UInt32(self, "error_correct_len")
        yield Bits(self, "stream_index", 7)
        yield Bits(self, "reserved[]", 8)
        yield Bit(self, "encrypted", "Content is encrypted?")
        yield UInt32(self, "reserved[]")
        size = self["data_len"].value
        if size:
            tag = self["type"].value
            if tag in Object.TAG_INFO:
                name, parser = Object.TAG_INFO[tag][0:2]
                yield parser(self, name, size=size * 8)
            else:
                yield RawBytes(self, "data", size)
        size = self["error_correct_len"].value
        if size:
            yield RawBytes(self, "error_correct", size)


class Object(FieldSet):
    # This list is converted to a dictionnary later where the key is the GUID
    TAG_INFO = (
        ("header", Header, "Header object"),
        ("file_prop", FileProperty, "File property"),
        ("header_ext", HeaderExtension, "Header extension"),
        ("codec_list", CodecList, "Codec list"),
        ("simple_index", SimpleIndex, "Simple index"),
        ("data", Data, "Data object"),
        ("stream_prop[]", StreamProperty, "Stream properties"),
        ("bit_rates", BitRateList, "Bit rate list"),
        ("ext_desc", ExtendedContentDescription, "Extended content description"),
        ("metadata", Metadata, "Metadata"),
        ("video_header", VideoHeader, "Video"),
        ("audio_header", AudioHeader, "Audio"),
        ("bitrate_mutex", BitrateMutualExclusion, "Bitrate mutual exclusion"),
    )

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)

        tag = self["guid"].value
        if tag not in self.TAG_INFO:
            self.handler = None
            return
        info = self.TAG_INFO[tag]
        self._name = info[0]
        self.handler = info[1]

    def createFields(self):
        yield GUID(self, "guid")
        yield filesizeHandler(UInt64(self, "size"))

        size = self["size"].value - self.current_size // 8
        if 0 < size:
            if self.handler:
                yield self.handler(self, "content", size=size * 8)
            else:
                yield RawBytes(self, "content", size)


tag_info_list = Object.TAG_INFO
Object.TAG_INFO = dict((parser[1].guid, parser) for parser in tag_info_list)


class AsfFile(Parser):
    MAGIC = b"\x30\x26\xB2\x75\x8E\x66\xCF\x11\xA6\xD9\x00\xAA\x00\x62\xCE\x6C"
    PARSER_TAGS = {
        "id": "asf",
        "category": "video",
        "file_ext": ("wmv", "wma", "asf"),
        "mime": ("video/x-ms-asf", "video/x-ms-wmv", "audio/x-ms-wma"),
        "min_size": 24 * 8,
        "description": "Advanced Streaming Format (ASF), used for WMV (video) and WMA (audio)",
        "magic": ((MAGIC, 0),),
    }
    FILE_TYPE = {
        "video/x-ms-wmv": (".wmv", "Window Media Video (wmv)"),
        "video/x-ms-asf": (".asf", "ASF container"),
        "audio/x-ms-wma": (".wma", "Window Media Audio (wma)"),
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        magic = self.MAGIC
        if self.stream.readBytes(0, len(magic)) != magic:
            return "Invalid magic"
        header = self[0]
        if not (30 <= header["size"].value <= MAX_HEADER_SIZE):
            return "Invalid header size (%u)" % header["size"].value
        return True

    def createMimeType(self):
        audio = False
        for prop in self.array("header/content/stream_prop"):
            guid = prop["content/type"].value
            if guid == VideoHeader.guid:
                return "video/x-ms-wmv"
            if guid == AudioHeader.guid:
                audio = True
        if audio:
            return "audio/x-ms-wma"
        else:
            return "video/x-ms-asf"

    def createFields(self):
        while not self.eof:
            yield Object(self, "object[]")

    def createDescription(self):
        return self.FILE_TYPE[self.mime_type][1]

    def createFilenameSuffix(self):
        return self.FILE_TYPE[self.mime_type][0]

    def createContentSize(self):
        if self[0].name != "header":
            return None
        return self["header/content/file_prop/content/file_size"].value * 8



from hachoir import field as field_module
from hachoir.wx.field_view.stubs import has_static_size, convert_size


def split_field(field, split_pos, split_name, split_t, size_func):
    split_name += '[]'

    subfields = [
        split_t(field._parent, split_name, split_pos),
        split_t(field._parent, split_name, size_func(field) - split_pos)]

    field._parent.replaceField(field.name, subfields)


def convert_field(field, new_type_name):
    field_set = field._parent
    new_type = getattr(field_module, new_type_name)

    if has_static_size(new_type):
        new_field = new_type(field_set, field.name, field._getDescription())
    else:
        new_field = new_type(field_set, field.name, convert_size(field, new_type), field._getDescription())

    field_set.replaceField(field.name, [new_field])



from hachoir.metadata.timezone import UTC
from datetime import date, datetime

# Year in 1850..2030
MIN_YEAR = 1850
MAX_YEAR = 2030


class Filter:

    def __init__(self, valid_types, min=None, max=None):
        self.types = valid_types
        self.min = min
        self.max = max

    def __call__(self, value):
        if not isinstance(value, self.types):
            return True
        if self.min is not None and value < self.min:
            return False
        if self.max is not None and self.max < value:
            return False
        return True


class NumberFilter(Filter):

    def __init__(self, min=None, max=None):
        Filter.__init__(self, (int, float), min, max)


class DatetimeFilter(Filter):

    def __init__(self, min=None, max=None):
        Filter.__init__(self, (date, datetime),
                        datetime(MIN_YEAR, 1, 1),
                        datetime(MAX_YEAR, 12, 31))
        self.min_date = date(MIN_YEAR, 1, 1)
        self.max_date = date(MAX_YEAR, 12, 31)
        self.min_tz = datetime(MIN_YEAR, 1, 1, tzinfo=UTC)
        self.max_tz = datetime(MAX_YEAR, 12, 31, tzinfo=UTC)

    def __call__(self, value):
        """
        Use different min/max values depending on value type
        (datetime with timezone, datetime or date).
        """
        if not isinstance(value, self.types):
            return True
        if hasattr(value, "tzinfo") and value.tzinfo:
            return (self.min_tz <= value <= self.max_tz)
        elif isinstance(value, datetime):
            return (self.min <= value <= self.max)
        else:
            return (self.min_date <= value <= self.max_date)


DATETIME_FILTER = DatetimeFilter()



# Field classes
from hachoir.field.field import Field, FieldError, MissingField, joinPath  # noqa
from hachoir.field.bit_field import Bit, Bits, RawBits  # noqa
from hachoir.field.byte_field import Bytes, RawBytes  # noqa
from hachoir.field.sub_file import SubFile, CompressedField  # noqa
from hachoir.field.character import Character  # noqa
from hachoir.field.integer import (Int8,  Int16,  Int24,  Int32,  Int64,  # noqa
                                   UInt8, UInt16, UInt24, UInt32, UInt64,
                                   GenericInteger)
from hachoir.field.enum import Enum  # noqa
from hachoir.field.string_field import (GenericString,  # noqa
                                        String, CString, UnixLine,
                                        PascalString8, PascalString16,
                                        PascalString32)
from hachoir.field.padding import (PaddingBits, PaddingBytes,  # noqa
                                   NullBits, NullBytes)

# Functions
from hachoir.field.helper import (isString, isInteger,  # noqa
                                  createPaddingField, createNullField,
                                  createRawField, writeIntoFile,
                                  createOrphanField)

# FieldSet classes
from hachoir.field.fake_array import FakeArray  # noqa
from hachoir.field.basic_field_set import (BasicFieldSet,  # noqa
                                           ParserError, MatchError)
from hachoir.field.generic_field_set import GenericFieldSet  # noqa
from hachoir.field.seekable_field_set import SeekableFieldSet, RootSeekableFieldSet  # noqa
from hachoir.field.field_set import FieldSet  # noqa
from hachoir.field.static_field_set import StaticFieldSet  # noqa
from hachoir.field.parser import Parser  # noqa
from hachoir.field.vector import GenericVector, UserVector  # noqa

# Complex types
from hachoir.field.float import Float32, Float64, Float80  # noqa
from hachoir.field.timestamp import (  # noqa
    GenericTimestamp,
    TimestampUnix32, TimestampUnix64, TimestampMac32, TimestampUUID60,
    TimestampWin64, TimedeltaMillisWin64,
    DateTimeMSDOS32, TimeDateMSDOS32, TimedeltaWin64)

# Special Field classes
from hachoir.field.link import Link, Fragment  # noqa
from hachoir.field.fragment import FragmentGroup, CustomFragment  # noqa

available_types = (Bit, Bits, RawBits,
                   Bytes, RawBytes,
                   SubFile,
                   Character,
                   Int8, Int16, Int24, Int32, Int64,
                   UInt8, UInt16, UInt24, UInt32, UInt64,
                   String, CString, UnixLine,
                   PascalString8, PascalString16, PascalString32,
                   Float32, Float64,
                   PaddingBits, PaddingBytes,
                   NullBits, NullBytes,
                   TimestampUnix32, TimestampMac32, TimestampWin64,
                   TimedeltaMillisWin64,
                   DateTimeMSDOS32, TimeDateMSDOS32,
                   #                   GenericInteger, GenericString,
                   )



"""
PCX picture filter.
"""

from hachoir.parser import Parser
from hachoir.field import (
    UInt8, UInt16,
    PaddingBytes, RawBytes,
    Enum)
from hachoir.parser.image.common import PaletteRGB
from hachoir.core.endian import LITTLE_ENDIAN


class PcxFile(Parser):
    endian = LITTLE_ENDIAN
    PARSER_TAGS = {
        "id": "pcx",
        "category": "image",
        "file_ext": ("pcx",),
        "mime": ("image/x-pcx",),
        "min_size": 128 * 8,
        "description": "PC Paintbrush (PCX) picture"
    }
    compression_name = {1: "Run-length encoding (RLE)"}
    version_name = {
        0: "Version 2.5 of PC Paintbrush",
        2: "Version 2.8 with palette information",
        3: "Version 2.8 without palette information",
        4: "PC Paintbrush for Windows",
        5: "Version 3.0 (or greater) of PC Paintbrush"
    }

    def validate(self):
        if self["id"].value != 10:
            return "Wrong signature"
        if self["version"].value not in self.version_name:
            return "Unknown format version"
        if self["bpp"].value not in (1, 2, 4, 8, 24, 32):
            return "Unknown bits/pixel"
        if self["reserved[0]"].value != b"\0":
            return "Invalid reserved value"
        return True

    def createFields(self):
        yield UInt8(self, "id", "PCX identifier (10)")
        yield Enum(UInt8(self, "version", "PCX version"), self.version_name)
        yield Enum(UInt8(self, "compression", "Compression method"), self.compression_name)
        yield UInt8(self, "bpp", "Bits / pixel")
        yield UInt16(self, "xmin", "Minimum X")
        yield UInt16(self, "ymin", "Minimum Y")
        yield UInt16(self, "xmax", "Width minus one")  # value + 1
        yield UInt16(self, "ymax", "Height minus one")  # value + 1
        yield UInt16(self, "horiz_dpi", "Horizontal DPI")
        yield UInt16(self, "vert_dpi", "Vertical DPI")
        yield PaletteRGB(self, "palette_4bits", 16, "Palette (4 bits)")
        yield PaddingBytes(self, "reserved[]", 1)
        yield UInt8(self, "nb_color_plan", "Number of color plans")
        yield UInt16(self, "bytes_per_line", "Bytes per line")
        yield UInt16(self, "color_mode", "Color mode")
        yield PaddingBytes(self, "reserved[]", 58)

        if self._size is None:  # TODO: is it possible to handle piped input?
            raise NotImplementedError

        nb_colors = 256
        size = (self._size - self.current_size) // 8
        has_palette = self["bpp"].value == 8
        if has_palette:
            size -= nb_colors * 3
        yield RawBytes(self, "image_data", size, "Image data")

        if has_palette:
            yield PaletteRGB(self, "palette_8bits", nb_colors, "Palette (8 bit)")



from hachoir.field import BasicFieldSet, GenericFieldSet, ParserError, createRawField


# getgaps(int, int, [listof (int, int)]) -> generator of (int, int)
# Gets all the gaps not covered by a block in `blocks` from `start` for
# `length` units.
def getgaps(start, length, blocks):
    '''
    Example:
    >>> list(getgaps(0, 20, [(15,3), (6,2), (6,2), (1,2), (2,3), (11,2), (9,5)]))
    [(0, 1), (5, 1), (8, 1), (14, 1), (18, 2)]
    '''
    # done this way to avoid mutating the original
    blocks = sorted(blocks, key=lambda b: b[0])
    end = start + length
    for s, l in blocks:
        if s > start:
            yield (start, s - start)
            start = s
        if s + l > start:
            start = s + l
    if start < end:
        yield (start, end - start)


class RootSeekableFieldSet(GenericFieldSet):

    def close(self):
        self.stream.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def seekBit(self, address, relative=True):
        if not relative:
            address -= self.absolute_address
        if address < 0:
            raise ParserError(
                "Seek below field set start (%s.%s)" % divmod(address, 8))
        self._current_size = address
        return None

    def seekByte(self, address, relative=True):
        return self.seekBit(address * 8, relative)

    def _fixLastField(self):
        """
        Try to fix last field when we know current field set size.
        Returns new added field if any, or None.
        """
        assert self._size is not None

        # Stop parser
        message = ["stop parser"]
        self._field_generator = None

        # If last field is too big, delete it
        while self._size < self._current_size:
            field = self._deleteField(len(self._fields) - 1)
            message.append("delete field %s" % field.path)
        assert self._current_size <= self._size

        blocks = [(x.absolute_address, x.size) for x in self._fields]
        fields = []
        self._size = max(self._size, max(
            a + b for a, b in blocks) - self.absolute_address)
        for start, length in getgaps(self.absolute_address, self._size, blocks):
            self.seekBit(start, relative=False)
            field = createRawField(self, length, "unparsed[]")
            self.setUniqueFieldName(field)
            self._fields.append(field.name, field)
            fields.append(field)
            message.append(
                "found unparsed segment: start %s, length %s" % (start, length))
        self.seekBit(self._size + self.absolute_address, relative=False)
        message = ", ".join(message)
        if fields:
            self.warning("[Autofix] Fix parser error: " + message)
        return fields

    def _stopFeeding(self):
        new_field = None
        if self._size is None:
            if self._parent:
                self._size = self._current_size

        new_field = self._fixLastField()
        self._field_generator = None
        return new_field


class SeekableFieldSet(RootSeekableFieldSet):

    def __init__(self, parent, name, description=None, size=None):
        assert issubclass(parent.__class__, BasicFieldSet)
        RootSeekableFieldSet.__init__(
            self, parent, name, parent.stream, description, size)



"""
Master Boot Record.


"""

# cfdisk uses the following algorithm to compute the geometry:
# 0. Use the values given by the user.
# 1. Try to guess the geometry from the partition table:
#    if all the used partitions end at the same head H and the
#    same sector S, then there are (H+1) heads and S sectors/cylinder.
# 2. Ask the system (ioctl/HDIO_GETGEO).
# 3. 255 heads and 63 sectors/cylinder.

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           Enum, Bits, UInt8, UInt16, UInt32,
                           RawBytes)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.tools import humanFilesize
from hachoir.core.text_handler import textHandler, hexadecimal

BLOCK_SIZE = 512  # bytes


class CylinderNumber(Bits):

    def __init__(self, parent, name, description=None):
        Bits.__init__(self, parent, name, 10, description)

    def createValue(self):
        i = self.parent.stream.readInteger(
            self.absolute_address, False, self._size, self.parent.endian)
        return i >> 2 | i % 4 << 8


class PartitionHeader(FieldSet):
    static_size = 16 * 8

    # taken from the source of cfdisk:
    # sed -n 's/.*{\(.*\), N_(\(.*\))}.*/        \1: \2,/p' i386_sys_types.c
    system_name = {
        0x00: "Empty",
        0x01: "FAT12",
        0x02: "XENIX root",
        0x03: "XENIX usr",
        0x04: "FAT16 <32M",
        0x05: "Extended",
        0x06: "FAT16",
        0x07: "HPFS/NTFS",
        0x08: "AIX",
        0x09: "AIX bootable",
        0x0a: "OS/2 Boot Manager",
        0x0b: "W95 FAT32",
        0x0c: "W95 FAT32 (LBA)",
        0x0e: "W95 FAT16 (LBA)",
        0x0f: "W95 Ext'd (LBA)",
        0x10: "OPUS",
        0x11: "Hidden FAT12",
        0x12: "Compaq diagnostics",
        0x14: "Hidden FAT16 <32M",
        0x16: "Hidden FAT16",
        0x17: "Hidden HPFS/NTFS",
        0x18: "AST SmartSleep",
        0x1b: "Hidden W95 FAT32",
        0x1c: "Hidden W95 FAT32 (LBA)",
        0x1e: "Hidden W95 FAT16 (LBA)",
        0x24: "NEC DOS",
        0x39: "Plan 9",
        0x3c: "PartitionMagic recovery",
        0x40: "Venix 80286",
        0x41: "PPC PReP Boot",
        0x42: "SFS",
        0x4d: "QNX4.x",
        0x4e: "QNX4.x 2nd part",
        0x4f: "QNX4.x 3rd part",
        0x50: "OnTrack DM",
        0x51: "OnTrack DM6 Aux1",
        0x52: "CP/M",
        0x53: "OnTrack DM6 Aux3",
        0x54: "OnTrackDM6",
        0x55: "EZ-Drive",
        0x56: "Golden Bow",
        0x5c: "Priam Edisk",
        0x61: "SpeedStor",
        0x63: "GNU HURD or SysV",
        0x64: "Novell Netware 286",
        0x65: "Novell Netware 386",
        0x70: "DiskSecure Multi-Boot",
        0x75: "PC/IX",
        0x80: "Old Minix",
        0x81: "Minix / old Linux",
        0x82: "Linux swap / Solaris",
        0x83: "Linux (ext2/ext3)",
        0x84: "OS/2 hidden C: drive",
        0x85: "Linux extended",
        0x86: "NTFS volume set",
        0x87: "NTFS volume set",
        0x88: "Linux plaintext",
        0x8e: "Linux LVM",
        0x93: "Amoeba",
        0x94: "Amoeba BBT",
        0x9f: "BSD/OS",
        0xa0: "IBM Thinkpad hibernation",
        0xa5: "FreeBSD",
        0xa6: "OpenBSD",
        0xa7: "NeXTSTEP",
        0xa8: "Darwin UFS",
        0xa9: "NetBSD",
        0xab: "Darwin boot",
        0xb7: "BSDI fs",
        0xb8: "BSDI swap",
        0xbb: "Boot Wizard hidden",
        0xbe: "Solaris boot",
        0xbf: "Solaris",
        0xc1: "DRDOS/sec (FAT-12)",
        0xc4: "DRDOS/sec (FAT-16 < 32M)",
        0xc6: "DRDOS/sec (FAT-16)",
        0xc7: "Syrinx",
        0xda: "Non-FS data",
        0xdb: "CP/M / CTOS / ...",
        0xde: "Dell Utility",
        0xdf: "BootIt",
        0xe1: "DOS access",
        0xe3: "DOS R/O",
        0xe4: "SpeedStor",
        0xeb: "BeOS fs",
        0xee: "EFI GPT",
        0xef: "EFI (FAT-12/16/32)",
        0xf0: "Linux/PA-RISC boot",
        0xf1: "SpeedStor",
        0xf4: "SpeedStor",
        0xf2: "DOS secondary",
        0xfd: "Linux raid autodetect",
        0xfe: "LANstep",
        0xff: "BBT"
    }

    def createFields(self):
        yield UInt8(self, "bootable", "Bootable flag (true if equals to 0x80)")
        if self["bootable"].value not in (0x00, 0x80):
            self.warning(
                "Stream doesn't look like master boot record (partition bootable error)!")
        yield UInt8(self, "start_head", "Starting head number of the partition")
        yield Bits(self, "start_sector", 6, "Starting sector number of the partition")
        yield CylinderNumber(self, "start_cylinder", "Starting cylinder number of the partition")
        yield Enum(UInt8(self, "system", "System indicator"), self.system_name)
        yield UInt8(self, "end_head", "Ending head number of the partition")
        yield Bits(self, "end_sector", 6, "Ending sector number of the partition")
        yield CylinderNumber(self, "end_cylinder", "Ending cylinder number of the partition")
        yield UInt32(self, "LBA", "LBA (number of sectors before this partition)")
        yield UInt32(self, "size", "Size (block count)")

    def isUsed(self):
        return self["system"].value != 0

    def createDescription(self):
        desc = "Partition header: "
        if self.isUsed():
            system = self["system"].display
            size = self["size"].value * BLOCK_SIZE
            desc += "%s, %s" % (system, humanFilesize(size))
        else:
            desc += "(unused)"
        return desc


class MasterBootRecord(FieldSet):
    static_size = 512 * 8

    def createFields(self):
        yield RawBytes(self, "program", 446, "Boot program (Intel x86 machine code)")
        yield PartitionHeader(self, "header[0]")
        yield PartitionHeader(self, "header[1]")
        yield PartitionHeader(self, "header[2]")
        yield PartitionHeader(self, "header[3]")
        yield textHandler(UInt16(self, "signature", "Signature (0xAA55)"), hexadecimal)

    def _getPartitions(self):
        return (self[index] for index in range(1, 5))
    headers = property(_getPartitions)


class Partition(FieldSet):

    def createFields(self):
        mbr = MasterBootRecord(self, "mbr")
        yield mbr

        # No error if we only want to analyse a backup of a mbr
        if self.eof:
            return

        for start, index, header in sorted((hdr["LBA"].value, index, hdr)
                                           for index, hdr in enumerate(mbr.headers) if hdr.isUsed()):
            # Seek to the beginning of the partition
            padding = self.seekByte(start * BLOCK_SIZE, "padding[]")
            if padding:
                yield padding

            # Content of the partition
            name = "partition[%u]" % index
            size = BLOCK_SIZE * header["size"].value
            desc = header["system"].display
            if header["system"].value == 5:
                yield Partition(self, name, desc, size * 8)
            else:
                yield RawBytes(self, name, size, desc)

        # Padding at the end
        if self.current_size < self._size:
            yield self.seekBit(self._size, "end")


class MSDos_HardDrive(Parser, Partition):
    endian = LITTLE_ENDIAN
    MAGIC = b"\x55\xAA"
    PARSER_TAGS = {
        "id": "msdos_harddrive",
        "category": "file_system",
        "description": "MS-DOS hard drive with Master Boot Record (MBR)",
        "min_size": 512 * 8,
        "file_ext": ("",),
        #        "magic": ((MAGIC, 510*8),),
    }

    def validate(self):
        if self.stream.readBytes(510 * 8, 2) != self.MAGIC:
            return "Invalid signature"
        used = False
        for hdr in self["mbr"].headers:
            if hdr["bootable"].value not in (0x00, 0x80):
                return "Wrong boot flag"
            used |= hdr.isUsed()
        return used or "No partition found"



from datetime import tzinfo, timedelta


class TimezoneUTC(tzinfo):
    """UTC timezone"""
    ZERO = timedelta(0)

    def utcoffset(self, dt):
        return TimezoneUTC.ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return TimezoneUTC.ZERO

    def __repr__(self):
        return "<TimezoneUTC delta=0, name='UTC'>"


class Timezone(TimezoneUTC):
    """Fixed offset in hour from UTC."""

    def __init__(self, offset):
        self._offset = timedelta(minutes=offset * 60)
        self._name = "%+03u00" % offset

    def utcoffset(self, dt):
        return self._offset

    def tzname(self, dt):
        return self._name

    def __repr__(self):
        return "<Timezone delta=%s, name='%s'>" % (
            self._offset, self._name)


UTC = TimezoneUTC()


def createTimezone(offset):
    if offset:
        return Timezone(offset)
    else:
        return UTC



"""
Very basic field: raw content with a size in byte. Use this class for
unknown content.
"""

from hachoir.field import Field, FieldError
from hachoir.core.tools import makePrintable
from hachoir.core import config

MAX_LENGTH = (2**64)


class RawBytes(Field):
    """
    Byte vector of unknown content

    @see: L{Bytes}
    """
    static_size = staticmethod(lambda *args, **kw: args[1] * 8)

    def __init__(self, parent, name, length, description="Raw data"):
        assert issubclass(parent.__class__, Field)
        if not (0 < length <= MAX_LENGTH):
            raise FieldError("Invalid RawBytes length (%s)!" % length)
        Field.__init__(self, parent, name, length * 8, description)
        self._display = None

    def _createDisplay(self, human):
        max_bytes = config.max_byte_length
        try:
            display = makePrintable(self.value[:max_bytes], "ASCII")
        except Exception:
            if self._display is None:
                address = self.absolute_address
                length = min(self._size // 8, max_bytes)
                self._display = self._parent.stream.readBytes(address, length)
            display = makePrintable(self._display, "ASCII")
        truncated = (8 * len(display) < self._size)
        if human:
            if truncated:
                display += "(...)"
            return makePrintable(display, "latin-1", quote='"')
        else:
            if truncated:
                return '"%s(...)"' % display
            else:
                return '"%s"' % display

    def createDisplay(self):
        return self._createDisplay(True)

    def createRawDisplay(self):
        return self._createDisplay(False)

    def hasValue(self):
        return True

    def createValue(self):
        assert (self._size % 8) == 0
        if self._display:
            self._display = None
        return self._parent.stream.readBytes(
            self.absolute_address, self._size // 8)


class Bytes(RawBytes):
    """
    Byte vector: can be used for magic number or GUID/UUID for example.

    @see: L{RawBytes}
    """
    pass



from wx import EVT_MENU
from wx.xrc import XRCID


class field_menu_fwd_t:
    def __init__(self, imp):
        self.imp = imp

    def on_field_view_ready(self, dispatcher, view):
        assert view is not None

        view.Bind(EVT_MENU, self.imp.on_addr_rel,
                  id=XRCID('field_menu_address_relative'))
        view.Bind(EVT_MENU, self.imp.on_addr_abs,
                  id=XRCID('field_menu_address_absolute'))
        view.Bind(EVT_MENU, self.imp.on_addr_hex,
                  id=XRCID('field_menu_address_base_hex'))
        view.Bind(EVT_MENU, self.imp.on_addr_dec,
                  id=XRCID('field_menu_address_base_dec'))
        view.Bind(EVT_MENU, self.imp.on_dump_to_disk,
                  id=XRCID('field_menu_dump_to_disk'))
        view.Bind(EVT_MENU, self.imp.on_parse_substream,
                  id=XRCID('field_menu_parse_substream'))
        view.Bind(EVT_MENU, self.imp.on_open_window_here,
                  id=XRCID('field_menu_open_window_here'))



from hachoir.metadata.metadata import RootMetadata, registerExtractor
from hachoir.metadata.safe import fault_tolerant
from hachoir.parser.container import SwfFile
from hachoir.parser.misc import TorrentFile, TrueTypeFontFile, OLE2_File, PcfFile
from hachoir.field import isString
from hachoir.core.error import warning
from hachoir.parser import guessParser
from hachoir.metadata.setter import normalizeString


class TorrentMetadata(RootMetadata):
    KEY_TO_ATTR = {
        "announce": "url",
        "comment": "comment",
        "creation_date": "creation_date",
    }
    INFO_TO_ATTR = {
        "length": "file_size",
        "name": "filename",
    }

    def extract(self, torrent):
        for field in torrent[0]:
            self.processRoot(field)

    @fault_tolerant
    def processRoot(self, field):
        if field.name in self.KEY_TO_ATTR:
            key = self.KEY_TO_ATTR[field.name]
            value = field.value
            setattr(self, key, value)
        elif field.name == "info" and "value" in field:
            for field in field["value"]:
                self.processInfo(field)

    @fault_tolerant
    def processInfo(self, field):
        if field.name in self.INFO_TO_ATTR:
            key = self.INFO_TO_ATTR[field.name]
            value = field.value
            setattr(self, key, value)
        elif field.name == "piece_length":
            self.comment = "Piece length: %s" % field.display


class TTF_Metadata(RootMetadata):
    NAMEID_TO_ATTR = {
        0: "copyright",   # Copyright notice
        3: "title",       # Unique font identifier
        5: "version",     # Version string
        8: "author",      # Manufacturer name
        11: "url",        # URL Vendor
        14: "copyright",  # License info URL
    }

    def extract(self, ttf):
        if "header" in ttf:
            self.extractHeader(ttf["header"])
        if "names" in ttf:
            self.extractNames(ttf["names"])

    @fault_tolerant
    def extractHeader(self, header):
        self.creation_date = header["created"].value
        self.last_modification = header["modified"].value
        self.comment = "Smallest readable size in pixels: %s pixels" % header[
            "lowest"].value
        self.comment = "Font direction: %s" % header["font_dir"].display

    @fault_tolerant
    def extractNames(self, names):
        offset = names["offset"].value
        for header in names.array("header"):
            key = header["nameID"].value
            foffset = offset + header["offset"].value
            field = names.getFieldByAddress(foffset * 8)
            if not field or not isString(field):
                continue
            value = field.value
            if key not in self.NAMEID_TO_ATTR:
                continue
            key = self.NAMEID_TO_ATTR[key]
            if key == "version" and value.startswith("Version "):
                # "Version 1.2" => "1.2"
                value = value[8:]
            setattr(self, key, value)


class OLE2_Metadata(RootMetadata):
    SUMMARY_ID_TO_ATTR = {
        2: "title",     # Title
        3: "title",     # Subject
        4: "author",
        6: "comment",
        8: "author",    # Last saved by
        12: "creation_date",
        13: "last_modification",
        14: "nb_page",
        18: "producer",
    }
    IGNORE_SUMMARY = set((
        1,  # Code page
    ))

    DOC_SUMMARY_ID_TO_ATTR = {
        3: "title",     # Subject
        14: "author",    # Manager
    }
    IGNORE_DOC_SUMMARY = set((
        1,  # Code page
    ))

    def extract(self, ole2):
        self._extract(ole2)

    def _extract(self, fieldset):
        try:
            fieldset._feedAll()
        except StopIteration:
            pass
        if "root[0]" in fieldset:
            self._extract(self.getFragment(fieldset["root[0]"]))
        doc_summary = self.getField(fieldset, "doc_summary[0]")
        if doc_summary:
            self.useSummary(doc_summary, True)
        word_doc = self.getField(fieldset, "word_doc[0]")
        if word_doc:
            self.useWordDocument(word_doc)
        summary = self.getField(fieldset, "summary[0]")
        if summary:
            self.useSummary(summary, False)

    def getFragment(self, frag):
        stream = frag.getSubIStream()
        ministream = guessParser(stream)
        if not ministream:
            warning("Unable to create the OLE2 mini stream parser!")
            return frag
        return ministream

    def getField(self, fieldset, name):
        # _feedAll() is needed to make sure that we get all fragments
        # eg. summary[0], summary[1], ..., summary[n]
        try:
            fieldset._feedAll()
        except StopIteration:
            pass
        if name not in fieldset:
            return None
        field = fieldset[name]
        return self.getFragment(field)

    @fault_tolerant
    def useSummary(self, summary, is_doc_summary):
        if "os" in summary:
            self.os = summary["os"].display
        if "section[0]" not in summary:
            return
        summary = summary["section[0]"]
        for property in summary.array("property_index"):
            self.useProperty(summary, property, is_doc_summary)

    @fault_tolerant
    def useWordDocument(self, doc):
        self.comment = "Encrypted: %s" % doc["FIB/fEncrypted"].value

    @fault_tolerant
    def useProperty(self, summary, property, is_doc_summary):
        field = summary.getFieldByAddress(property["offset"].value * 8)
        if not field \
                or "value" not in field:
            return
        field = field["value"]
        if not field.hasValue():
            return

        # Get value
        value = field.value
        if isinstance(value, str):
            value = normalizeString(value)
            if not value:
                return

        # Get property identifier
        prop_id = property["id"].value
        if is_doc_summary:
            id_to_attr = self.DOC_SUMMARY_ID_TO_ATTR
            ignore = self.IGNORE_DOC_SUMMARY
        else:
            id_to_attr = self.SUMMARY_ID_TO_ATTR
            ignore = self.IGNORE_SUMMARY
        if prop_id in ignore:
            return

        # Get Hachoir metadata key
        try:
            key = id_to_attr[prop_id]
            use_prefix = False
        except LookupError:
            key = "comment"
            use_prefix = True
        if use_prefix:
            prefix = property["id"].display
            if (prefix in ("TotalEditingTime", "LastPrinted")) \
                    and (not field):
                # Ignore null time delta
                return
            value = "%s: %s" % (prefix, value)
        else:
            if (key == "last_modification") and (not field):
                # Ignore null timestamp
                return
        setattr(self, key, value)


class PcfMetadata(RootMetadata):
    PROP_TO_KEY = {
        'CHARSET_REGISTRY': 'charset',
        'COPYRIGHT': 'copyright',
        'WEIGHT_NAME': 'font_weight',
        'FOUNDRY': 'author',
        'FONT': 'title',
        '_XMBDFED_INFO': 'producer',
    }

    def extract(self, pcf):
        if "properties" in pcf:
            self.useProperties(pcf["properties"])

    def useProperties(self, properties):
        last = properties["total_str_length"]
        offset0 = last.address + last.size
        for index in properties.array("property"):
            # Search name and value
            value = properties.getFieldByAddress(
                offset0 + index["value_offset"].value * 8)
            if not value:
                continue
            value = value.value
            if not value:
                continue
            name = properties.getFieldByAddress(
                offset0 + index["name_offset"].value * 8)
            if not name:
                continue
            name = name.value
            if name not in self.PROP_TO_KEY:
                warning("Skip %s=%r" % (name, value))
                continue
            key = self.PROP_TO_KEY[name]
            setattr(self, key, value)


class SwfMetadata(RootMetadata):

    def extract(self, swf):
        self.height = swf["rect/ymax"].value  # twips
        self.width = swf["rect/xmax"].value  # twips
        self.format_version = "flash version %s" % swf["version"].value
        self.frame_rate = swf["frame_rate"].value
        self.comment = "Frame count: %s" % swf["frame_count"].value


registerExtractor(TorrentFile, TorrentMetadata)
registerExtractor(TrueTypeFontFile, TTF_Metadata)
registerExtractor(OLE2_File, OLE2_Metadata)
registerExtractor(PcfFile, PcfMetadata)
registerExtractor(SwfFile, SwfMetadata)



"""
Parser for an obscure FMV file format: bin files from the game
"The Amazing Spider-Man vs. The Kingpin" (Sega CD)

Author: Mike Melanson
Creation date: 2006-09-30
File samples: http://samples.mplayerhq.hu/game-formats/spiderman-segacd-bin/
"""

from hachoir.parser import Parser
from hachoir.field import FieldSet, UInt32, String, RawBytes
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal


class Chunk(FieldSet):
    tag_info = {
        "CONF": ("conf[]", None, "Configuration header"),
        "AUDI": ("audio[]", None, "Audio chunk"),
        "SYNC": ("sync[]", None, "Start of video frame data"),
        "IVRA": ("ivra[]", None, "Vector codebook (?)"),
        "VRAM": ("video[]", None, "Video RAM tile pattern"),
        "CRAM": ("color[]", None, "Color RAM (palette)"),
        "CEND": ("video_end[]", None, "End of video data"),
        "MEND": ("end_file", None, "End of file"),
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = self["length"].value * 8
        fourcc = self["fourcc"].value
        if fourcc in self.tag_info:
            self._name, self._parser, self._description = self.tag_info[fourcc]
        else:
            self._parser = None
            self._description = "Unknown chunk: fourcc %s" % self[
                "fourcc"].display

    def createFields(self):
        yield String(self, "fourcc", 4, "FourCC", charset="ASCII")
        yield textHandler(UInt32(self, "length", "length"), hexadecimal)
        size = self["length"].value - 8
        if 0 < size:
            if self._parser:
                yield from self._parser(self, size)
            else:
                yield RawBytes(self, "data", size)


class SpiderManVideoFile(Parser):
    PARSER_TAGS = {
        "id": "spiderman_video",
        "category": "game",
        "file_ext": ("bin",),
        "min_size": 8 * 8,
        "description": "The Amazing Spider-Man vs. The Kingpin (Sega CD) FMV video"
    }

    endian = BIG_ENDIAN

    def validate(self):
        return (self.stream.readBytes(0, 4) == b'CONF')

    def createFields(self):
        while not self.eof:
            yield Chunk(self, "chunk[]")



# -*- coding: utf-8 -*-

from hachoir.wx.dialogs import file_save_dialog
import wx


class field_menu_t:
    def __init__(self, parent, menu):
        self.parent = parent
        self.menu = menu

        # forward this call because xrc doesn't allow menu
        # subclassing (as of 2.6.3)
        self.Bind = self.menu.Bind

    def show_opts(self):
        self.parent.PopupMenu(self.menu)

    def ask_for_dump_file(self, title):
        dump_dlog = file_save_dialog(title)
        if wx.ID_OK == dump_dlog.ShowModal():
            return dump_dlog.GetPath()



"""
Truevision Targa Graphic (TGA) picture parser.

Author: Victor Stinner
Creation: 18 december 2006
"""

from hachoir.parser import Parser
from hachoir.field import FieldSet, UInt8, UInt16, Enum, RawBytes, Bit, Bits, RawBits
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.parser.image.common import PaletteRGB
from hachoir.core.text_handler import textHandler, hexadecimal


class Line(FieldSet):

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = self["/width"].value * self.root.getBpp()

    def createFields(self):
        bpp = self.root.getBpp()
        for x in range(self["/width"].value):
            yield textHandler(Bits(self, "pixel[]", bpp), hexadecimal)


class Pixels(FieldSet):

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = self["/width"].value * \
            self["/height"].value * self.root.getBpp()

    def createFields(self):
        if self["/y_flip"].value:
            RANGE = range(self["/height"].value)
        else:
            RANGE = reversed(range(self["/height"].value))
        for y in RANGE:
            yield Line(self, "line[%u]" % y)


class TargaFile(Parser):
    PARSER_TAGS = {
        "id": "targa",
        "category": "image",
        "file_ext": ("tga",),
        "mime": ("image/targa", "image/tga", "image/x-tga"),
        "min_size": 18 * 8,
        "description": "Truevision Targa Graphic (TGA)"
    }
    CODEC_NAME = {
        0: "No image data",
        1: "Palette uncompressed",
        2: "True-color uncompressed",
        3: "Grayscale uncompressed",
        9: "Palette RLE",
        10: "True-color RLE",
        11: "Grayscale RLE",
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        if self["codec"].value not in self.CODEC_NAME:
            return "Unknown codec"
        if self["palette_type"].value not in (0, 1):
            return "Unknown palette type %d" % self["palette_type"].value
        if self["bpp"].value not in (8, 15, 16, 24, 32):
            return "Unknown bits/pixel value %d" % self["bpp"].value
        return True

    def getBpp(self):
        bpp = self['bpp'].value
        if bpp == 15:
            bpp = 16
        return bpp

    def createFields(self):
        yield UInt8(self, "id_length", "Length of the image ID field")
        yield UInt8(self, "palette_type", "Colormap present?")
        yield Enum(UInt8(self, "codec", "Pixels encoding"), self.CODEC_NAME)
        yield UInt16(self, "palette_ofs", "Palette absolute file offset")
        yield UInt16(self, "nb_color", "Number of colors in the palette")
        yield UInt8(self, "color_map_size", "Size of each palette entry")
        yield UInt16(self, "x_min")
        yield UInt16(self, "y_min")
        yield UInt16(self, "width")
        yield UInt16(self, "height")
        yield UInt8(self, "bpp", "Bits per pixel")
        yield Bits(self, "alpha_depth", 4, "Alpha channel depth")
        yield Bit(self, "x_flip", "Flip across the X-axis? (If set, columns run right-to-left)")
        yield Bit(self, "y_flip", "Flip across the Y-axis? (If set, rows run top-to-bottom)")
        yield RawBits(self, "reserved_flags", 2)

        if self["id_length"].value:
            yield RawBytes(self, "image_id", self["id_length"].value)

        if self["palette_type"].value == 1:
            yield PaletteRGB(self, "palette", 1 << self["bpp"].value)

        if self["codec"] in (1, 2, 3):
            yield Pixels(self, "pixels")
        else:
            size = (self.size - self.current_size) // 8
            if size:
                yield RawBytes(self, "raw_pixels", size)



"""
IPTC metadata parser (can be found in a JPEG picture for example)

Sources:
- Image-MetaData Perl module:
  http://www.annocpan.org/~BETTELLI/Image-MetaData-JPEG-0.15/...
  ...lib/Image/MetaData/JPEG/TagLists.pod
- IPTC tag name and description:
  http://peccatte.karefil.com/software/IPTCTableau.pdf

Author: Victor Stinner
"""

from hachoir.field import (FieldSet, ParserError,
                           UInt8, UInt16, String, RawBytes, NullBytes)
from hachoir.core.text_handler import textHandler, hexadecimal


def IPTC_String(parent, name, desc=None):
    # Charset may be utf-8, ISO-8859-1, or ...
    return String(parent, name, parent["size"].value, desc,
                  strip=" ")


dataset1 = {
}
dataset2 = {
    0: ("record_version", "Record version (2 for JPEG)", UInt16),
    5: ("obj_name", "Object name", None),
    7: ("edit_stat", "Edit status", None),
    10: ("urgency", "Urgency", UInt8),
    15: ("category[]", "Category", None),
    22: ("fixture", "Fixture identifier", IPTC_String),
    25: ("keyword[]", "Keywords", IPTC_String),
    30: ("release_date", "Release date", IPTC_String),
    35: ("release_time", "Release time", IPTC_String),
    40: ("instruction", "Special instructions", IPTC_String),
    55: ("date_created", "Date created", IPTC_String),
    60: ("time_created", "Time created (ISO 8601)", IPTC_String),
    65: ("originating_prog", "Originating program", IPTC_String),
    70: ("prog_ver", "Program version", IPTC_String),
    80: ("author", "By-line (Author)", IPTC_String),
    85: ("author_job", "By-line (Author precision)", IPTC_String),
    90: ("city", "City", IPTC_String),
    95: ("state", "Province / State", IPTC_String),
    100: ("country_code", "Country / Primary location code", IPTC_String),
    101: ("country_name", "Country / Primary location name", IPTC_String),
    103: ("trans_ref", "Original transmission reference", IPTC_String),
    105: ("headline", "Headline", IPTC_String),
    110: ("credit", "Credit", IPTC_String),
    115: ("source", "Source", IPTC_String),
    116: ("copyright", "Copyright notice", IPTC_String),
    120: ("caption", "Caption/Abstract", IPTC_String),
    122: ("writer", "Writer/editor", IPTC_String),
    231: ("history[]", "Document history (timestamp)", IPTC_String)
}
datasets = {1: dataset1, 2: dataset2}


class IPTC_Size(FieldSet):

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        value = 0
        for field in self:
            value <<= 15
            value += (field.value & 0x7fff)
        self.createValue = lambda: value

    def createFields(self):
        while True:
            field = UInt16(self, "value[]")
            yield field
            if field.value < 0x8000:
                break


class IPTC_Chunk(FieldSet):

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        number = self["dataset_nb"].value
        self.dataset_info = None
        if number in datasets:
            tag = self["tag"].value
            if tag in datasets[number]:
                self.dataset_info = datasets[number][tag]
                self._name = self.dataset_info[0]
                self._description = self.dataset_info[1]
        size_chunk = self["size"]
        self._size = 3 * 8 + size_chunk.size + size_chunk.value * 8

    def createFields(self):
        yield textHandler(UInt8(self, "signature", "IPTC signature (0x1c)"), hexadecimal)
        if self["signature"].value != 0x1C:
            raise ParserError("Wrong IPTC signature")
        yield textHandler(UInt8(self, "dataset_nb", "Dataset number"), hexadecimal)
        yield UInt8(self, "tag", "Tag")
        yield IPTC_Size(self, "size", "Content size")

        size = self["size"].value
        if 0 < size:
            if self.dataset_info:
                cls = self.dataset_info[2]
            else:
                cls = None
            if cls:
                yield cls(self, "content")
            else:
                yield RawBytes(self, "content", size)


class IPTC(FieldSet):

    def createFields(self):
        while 5 <= (self._size - self.current_size) // 8:
            yield IPTC_Chunk(self, "chunk[]")
        size = (self._size - self.current_size) // 8
        if 0 < size:
            yield NullBytes(self, "padding", size)



from hachoir.field import (FieldSet,
                           UInt16, UInt32, Enum, String, Bytes, Bits, TimestampUUID60)
from hachoir.parser.video.fourcc import video_fourcc_name
from hachoir.core.bits import str2hex
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.parser.network.common import MAC48_Address

# Dictionary: Windows codepage => Python charset name
CODEPAGE_CHARSET = {
    874: "CP874",
    #     932: Japanese Shift-JIS
    #     936: Simplified Chinese GBK
    #     949: Korean
    #     950: Traditional Chinese Big5
    1250: "WINDOWS-1250",
    1251: "WINDOWS-1251",
    1252: "WINDOWS-1252",
    1253: "WINDOWS-1253",
    1254: "WINDOWS-1254",
    1255: "WINDOWS-1255",
    1256: "WINDOWS-1256",
    1257: "WINDOWS-1257",
    1258: "WINDOWS-1258",
    10000: "MacRoman",
    65001: "UTF-8",
}


class PascalStringWin16(FieldSet):

    def __init__(self, parent, name, description=None, strip=None, charset="UTF-16-LE"):
        FieldSet.__init__(self, parent, name, description)
        length = self["length"].value
        self._size = 16 + length * 16
        self.strip = strip
        self.charset = charset

    def createFields(self):
        yield UInt16(self, "length", "Length in widechar characters")
        size = self["length"].value
        if size:
            yield String(self, "text", size * 2, charset=self.charset, strip=self.strip)

    def createValue(self):
        if "text" in self:
            return self["text"].value
        else:
            return None


class PascalStringWin32(FieldSet):

    def __init__(self, parent, name, description=None, strip=None, charset="UTF-16-LE"):
        FieldSet.__init__(self, parent, name, description)
        length = self["length"].value
        self._size = 32 + length * 16
        self.strip = strip
        self.charset = charset

    def createFields(self):
        yield UInt32(self, "length", "Length in widechar characters")
        size = self["length"].value
        if size:
            yield String(self, "text", size * 2, charset=self.charset, strip=self.strip)

    def createValue(self):
        if "text" in self:
            return self["text"].value
        else:
            return None


class GUID(FieldSet):
    """
    Windows 128 bits Globally Unique Identifier (GUID)

    See RFC 4122
    """
    static_size = 128
    NULL = "00000000-0000-0000-0000-000000000000"
    FIELD_NAMES = {
        3: ("sha1_high", "sha1_low"),
        4: ("random_high", "random_low"),
        5: ("md5_high", "md5_low"),
    }
    VERSION_NAME = {
        1: "Timestamp & MAC-48",
        2: "DCE Security version",
        3: "Name SHA-1 hash",
        4: "Randomly generated",
        5: "Name MD5 hash",
    }
    VARIANT_NAME = {
        0: "NCS",
        2: "Leach-Salz",
        # 5: Microsoft Corporation?
        6: "Microsoft Corporation",
        7: "Reserved Future",
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self.version = self.stream.readBits(
            self.absolute_address + 32 + 16 + 12, 4, self.endian)

    def createFields(self):
        if self.version == 1:
            yield TimestampUUID60(self, "time")
            yield Enum(Bits(self, "version", 4), self.VERSION_NAME)
            yield Enum(Bits(self, "variant", 3), self.VARIANT_NAME)
            yield textHandler(Bits(self, "clock", 13), hexadecimal)
#            yield textHandler(Bits(self, "clock", 16), hexadecimal)
            if self.version == 1:
                yield MAC48_Address(self, "mac", "IEEE 802 MAC address")
            else:
                yield Bytes(self, "node", 6)
        else:
            namea, nameb = self.FIELD_NAMES.get(
                self.version, ("data_a", "data_b"))
            yield textHandler(Bits(self, namea, 60), hexadecimal)
            yield Enum(Bits(self, "version", 4), self.VERSION_NAME)
            yield Enum(Bits(self, "variant", 3), self.VARIANT_NAME)
            yield textHandler(Bits(self, nameb, 61), hexadecimal)

    def createValue(self):
        addr = self.absolute_address
        a = self.stream.readBits(addr, 32, self.endian)
        b = self.stream.readBits(addr + 32, 16, self.endian)
        c = self.stream.readBits(addr + 48, 16, self.endian)
        d = self.stream.readBytes(addr + 64, 2)
        e = self.stream.readBytes(addr + 80, 6)
        return "%08X-%04X-%04X-%s-%s" % (a, b, c, str2hex(d), str2hex(e))

    def createDisplay(self):
        value = self.value
        if value == self.NULL:
            name = "Null GUID: "
        else:
            name = "GUID v%u (%s): " % (self.version, self["version"].display)
        return name + value

    def createRawDisplay(self):
        value = self.stream.readBytes(self.absolute_address, 16)
        return str2hex(value, format=r"\x%02x")


class BitmapInfoHeader(FieldSet):
    """ Win32 BITMAPINFOHEADER structure from GDI """
    static_size = 40 * 8

    COMPRESSION_NAME = {
        0: "Uncompressed (RGB)",
        1: "RLE (8 bits)",
        2: "RLE (4 bits)",
        3: "Bitfields",
        4: "JPEG",
        5: "PNG"
    }

    def __init__(self, parent, name, use_fourcc=False):
        FieldSet.__init__(self, parent, name)
        self._use_fourcc = use_fourcc

    def createFields(self):
        yield UInt32(self, "hdr_size", "Header size (in bytes) (=40)")
        yield UInt32(self, "width", "Width")
        yield UInt32(self, "height", "Height")
        yield UInt16(self, "nb_planes", "Color planes")
        yield UInt16(self, "bpp", "Bits/pixel")
        if self._use_fourcc:
            yield Enum(String(self, "codec", 4, charset="ASCII"), video_fourcc_name)
        else:
            yield Enum(UInt32(self, "codec", "Compression"), self.COMPRESSION_NAME)
        yield UInt32(self, "size", "Image size (in bytes)")
        yield UInt32(self, "xres", "X pixels per meter")
        yield UInt32(self, "yres", "Y pixels per meter")
        yield UInt32(self, "color_used", "Number of used colors")
        yield UInt32(self, "color_important", "Number of important colors")

    def createDescription(self):
        return "Bitmap info header: %ux%u pixels, %u bits/pixel" % \
            (self["width"].value, self["height"].value, self["bpp"].value)



import wx


class field_split_menu_t:
    def __init__(self, parent, menu):
        self.parent = parent
        self.menu = menu
        self.Bind = self.menu.Bind   # see note in field_menu.py

    def ask_split(self, caption, min, max):
        # Note: we would prefer a NumberEntryDialog but this isn't currently wrapped
        # by wxPython Phoenix.
        res = None
        dlg = wx.TextEntryDialog(self.parent, 'Enter split offset:', '',
                                 caption, min, min, max)
        if dlg.ShowModal() == wx.ID_OK:
            try:
                res = int(dlg.GetValue())
            except ValueError:
                res = None
        dlg.Destroy()
        return res
